from pathlib import Path
import os, requests
import re
from glob import glob

from .parse import Iarg, parse_cmd, pat_to_regex, clean_data
from .consts import COOKIE_FILE, IARG_EMPTY, MONTH, YEAR, DAY, START_YEAR
from loguru import logger


def download_day(
    day: int,
    year: int,
    cookie: str,
    pattern: str,
    files: tuple[str, ...],
    file_pattern: str,
) -> None:
    folder = Path(pattern.format(day=day, year=year))
    logger.info(f"Downloading data for Day {day} of {year} into folder: {folder}")
    logger.debug(f"Target folder: {folder.absolute()}")
    os.makedirs(folder, exist_ok=True)

    url = f"https://adventofcode.com/{year}/day/{day}"
    response = requests.get(url)
    logger.debug(f"Request to {url} returned status code {response.status_code}")
    if response.status_code == 200:
        data = response.text
        if not (folder / "test.txt").exists():
            (folder / "test.txt").write_text(
                clean_data(
                    data.split("<pre><code>", maxsplit=1)[1].split(
                        "</code></pre>", maxsplit=1
                    )[0]
                )
            )
    else:
        logger.warning(
            f"Failed to download test data from {url}. This can happen if the day contains weird test data."
        )

    url += "/input"
    response = requests.get(url, cookies={"session": cookie})
    logger.debug(f"Request to {url} returned status code {response.status_code}")

    if response.status_code == 200:
        if not (folder / "input.txt").exists():
            (folder / "input.txt").write_text(response.text.rstrip())
    else:
        logger.error(
            f"Failed to download data from {url}. This usually means the cookie is invalid."
        )

    for file in filter(None, files):  # Skip empty filenames
        path = folder / file
        if not path.exists():
            logger.debug(f"Creating file: {path}")
            path.write_text(file_pattern)


def main():
    args = parse_cmd()

    if args.write_cookie is not None:
        COOKIE_FILE.write_text(args.write_cookie.strip())
        logger.success(f"Wrote new cookie to {COOKIE_FILE}")
        exit()

    if not args.cookie.exists():
        logger.error(
            f"Please create a cookie.txt file in the same directory as this script ({COOKIE_FILE})."
        )
        exit()

    COOKIE = args.cookie.read_text().strip()

    year: Iarg = args.year if args.year is not None else YEAR
    day: Iarg = args.day if args.day is not None else DAY
    PATTERN: str = args.pattern or ("{year}/Day_{day}" if year != YEAR else "Day_{day}")
    FILES: tuple[str, ...] = tuple(args.files)
    file_pattern: str = args.template
    logger.debug(f"year={year}, day={day}, pattern={PATTERN}, files={FILES}, template={file_pattern.replace('\n', '\\n')}")

    # If match, ignore all else
    if args.match is not None:
        glob_pattern = re.sub(r"\{.*\}", "*", args.match)
        compiled_pattern = pat_to_regex(args.match)
        logger.info(
            f"Matching folders with pattern: {args.match} -> {compiled_pattern.pattern}"
        )
        for folder in glob(glob_pattern):
            logger.info(f"Checking folder: {folder}")
            if match := compiled_pattern.match(folder):
                dct = match.groupdict()
                year = int(dct.get("year", YEAR))
                day = int(dct.get("day", DAY))
                download_day(
                    day,
                    year,
                    COOKIE,
                    pattern=folder,
                    files=FILES,
                    file_pattern=file_pattern,
                )
                logger.info(f"Downloaded Day {day} of {year}")
        exit()

    def to_list(
        value: int | tuple[int, int], default: list[int], name: str
    ) -> list[int]:
        if isinstance(value, tuple):
            s, e = value
            if s > e:
                print(f"Invalid {name} range {s}-{e}")
                exit()
            return list(range(s, e + 1))
        return default if value == IARG_EMPTY else [value]

    years = to_list(year, list(range(START_YEAR, YEAR + 1)), "year")
    days = to_list(day, list(range(1, max(26, DAY + 1))), "day")

    for year in years:
        if year < START_YEAR or year > YEAR:
            logger.warning(f"Year {year} is out of valid range ({START_YEAR}-{YEAR})")
            continue
        for day in days:
            if year == YEAR and MONTH == 12 and day > DAY:
                logger.warning(f"Day {day} of this year ({year}) has not arrived yet.")
                break
            download_day(
                day,
                year,
                COOKIE,
                pattern=PATTERN,
                files=FILES,
                file_pattern=file_pattern,
            )
            logger.info(f"Downloaded Day {day} of {year}")



import functools
import itertools
from typing import Iterator, Callable, Self


class Iter[T](Iterator[T]):
    __slots__ = ("iterable",)

    def __init__(self, iterable: Iterator[T]):
        self.iterable = iterable

    def __iter__(self) -> Iterator[T]:
        return self.iterable

    def __next__(self) -> T:
        return next(self.iterable)

    def filter(self, func) -> Self:
        self.iterable = filter(func, self.iterable)
        return self

    def map[U](self, func: Callable[[T], U]) -> "Iter[U]":
        return Iter(map(func, self.iterable))

    def flat_map[U](self, func: Callable[[T], Iterator[U]]) -> "Iter[U]":
        return Iter(y for x in self.iterable for y in func(x))

    def collect[U](self, t: Callable[[Iterator[T]], U] = list) -> U:
        return t(self.iterable)

    def any(self, func: Callable[[T], bool]) -> bool:
        return any(map(func, self.iterable))

    def all(self, func: Callable[[T], bool]) -> bool:
        return all(map(func, self.iterable))

    def find(self, func: Callable[[T], bool]) -> T:
        return next(filter(func, self.iterable))

    def try_find(self, func: Callable[[T], bool]) -> T | None:
        return next(filter(func, self.iterable), None)

    def fold(self, func: Callable[[T, T], T]) -> T:
        return functools.reduce(func, self.iterable)

    def ilen(self) -> int:
        return sum(1 for _ in self.iterable)

    def enumerate(self, start: int = 0) -> "Iter[tuple[int, T]]":
        return Iter(enumerate(self.iterable, start=start))

    def skip(self, n: int) -> Self:
        for _ in range(n):
            next(self.iterable, None)
        return self

    def take(self, n: int) -> Self:
        self.iterable = itertools.islice(self.iterable, n)
        return self

    def take_while(self, func: Callable[[T], bool]) -> Self:
        self.iterable = itertools.takewhile(func, self.iterable)
        return self

    def drop_while(self, func: Callable[[T], bool]) -> Self:
        self.iterable = itertools.dropwhile(func, self.iterable)
        return self

    def chain(self, other: Iterator[T]) -> Self:
        self.iterable = itertools.chain(self.iterable, other)
        return self



from datetime import datetime
from pathlib import Path

HOME = Path(__file__).parent
NOW = datetime.now()
DAY = NOW.day
MONTH = NOW.month
YEAR = NOW.year if MONTH >= 12 else NOW.year - 1 # Advent of Code runs in December
COOKIE_FILE = HOME / "cookie.txt"
START_YEAR = 2015
IARG_EMPTY = -1

FILE_TEMPLATE = """\
from pathlib import Path

HOME = Path(__file__).parent

with open(HOME/"test.txt") as f:
    pass
"""



class It_:
    __slots__ = ("f",)

    def __init__(self, f=lambda x: x) -> None:
        self.f = f

    def __call__(self, x):
        return self.f(x)

    def __lt__(self, other) -> "It_":
        return It_(lambda x: self.f(x) < other)

    def __gt__(self, other) -> "It_":
        return It_(lambda x: self.f(x) > other)

    def __le__(self, other) -> "It_":
        return It_(lambda x: self.f(x) <= other)

    def __ge__(self, other) -> "It_":
        return It_(lambda x: self.f(x) >= other)

    def __matmul__(self, other: "It_") -> "It_":
        return It_(lambda x: self.f(other.f(x)))
    
    def __eq__(self, other): # type: ignore
        return It_(lambda x: self.f(x) == other)
    
    def __ne__(self, other): # type: ignore
        return It_(lambda x: self.f(x) != other)
    
    def __or__(self, other: "It_") -> "It_":
        return It_(lambda x: self.f(x) or other.f(x))
    
    def __and__(self, other: "It_") -> "It_":
        return It_(lambda x: self.f(x) and other.f(x))
    
    def __sub__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) - other)
    
    def __add__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) + other)
    
    def __mul__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) * other)
    
    def __truediv__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) / other)
    
    def __floordiv__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) // other)
    
    def __mod__(self, other) -> "It_":
        if isinstance(other, It_):
            return It_(lambda x: self.f(x) - other.f(x))
        return It_(lambda x: self.f(x) % other)


It = It_()



from argparse import ArgumentParser, ArgumentTypeError, Namespace
from pathlib import Path
import html, re

from loguru import logger
from .consts import COOKIE_FILE, FILE_TEMPLATE, IARG_EMPTY

type Iarg = int | tuple[int, int]


def valid_iarg(value) -> Iarg:
    if value.isdigit():
        return int(value)
    if value == "all":
        return IARG_EMPTY
    if (
        "-" in value
        and len(parts := value.split("-")) == 2
        and all(map(str.isdigit, parts))
    ):
        return tuple(map(int, parts))  # type: ignore
    raise ArgumentTypeError(
        f"Invalid value: {value}. Must be a positive integer,"
        " two dash-separated positive integers, or 'all'."
    )

def parse_cmd() -> Namespace:
    parser = ArgumentParser()
    parser.add_argument("-d", "--day", type=valid_iarg, default=None, help="The day")
    parser.add_argument("-y", "--year", type=valid_iarg, default=None, help="The year")
    parser.add_argument(
        "-c", "--cookie", type=Path, default=COOKIE_FILE, help="The cookie file"
    )
    parser.add_argument(
        "-p",
        "--pattern",
        type=str,
        default=None,
        help="The pattern to use for the folder name",
    )
    parser.add_argument(
        "-m",
        "--match",
        type=str,
        default=None,
        help="Match folder structure and fill days that have no data",
    )
    parser.add_argument(
        "--write-cookie",
        type=str,
        default=None,
        help="Write a new cookie string to the cookie file",
    )
    parser.add_argument(
        "-f",
        "--files",
        type=str,
        nargs="+",
        default=("p1.py", "p2.py"),
        help="Specify filenames",
    )
    parser.add_argument(
        "-t",
        "--template",
        type=str,
        default=FILE_TEMPLATE,
        help="The template to use for new problem files",
    )
    
    return parser.parse_args()


def pat_to_regex(pattern) -> re.Pattern:
    escaped_pattern = re.escape(pattern).replace(r"\*", ".*")
    logger.debug(f"Escaped pattern: {pattern} -> {escaped_pattern}")
    named_regex_pattern = re.sub(
        r"\\\{(\w+).*\\\}", r"(?P<\1>\\d+)", escaped_pattern
    )
    logger.debug(f"Converted pattern: {pattern} -> {named_regex_pattern}")
    return re.compile(named_regex_pattern)


def clean_data(data: str) -> str:
    return html.unescape(data.removesuffix("\n"))
